/*
 * Copyright (c) 2010
 *	Ben Gray <ben.r.gray@gmail.com>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Ben Gray.
 * 4. The name of the company nor the name of the author may be used to
 *    endorse or promote products derived from this software without specific
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY BEN GRAY ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL BEN GRAY BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * Power, Reset and Clock Managment Module
 *
 * This is a very simple driver wrapper around the PRCM set of registers in
 * the OMAP3 chip. It allows you to turn on and off things like the functional
 * and interface clocks to the various on-chip modules.
 *
 */
#include <sys/cdefs.h>
__FBSDID("$FreeBSD$");

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/resource.h>
#include <sys/rman.h>
#include <sys/lock.h>
#include <sys/mutex.h>

#include <machine/bus.h>
#include <machine/cpu.h>
#include <machine/cpufunc.h>
#include <machine/frame.h>
#include <machine/resource.h>
#include <machine/intr.h>

#include <arm/omap/omap_prcm.h>

/**
 *	Structure that stores the driver context.
 *
 *	This structure is allocated during driver attach, it is not designed to be
 *	deallocated and a pointer to it is stored globally (g_omap3_prcm_softc).
 */
struct omap_prcm_softc {
	device_t		sc_dev;
	
	/* 
	 * The memory resource(s) for the PRCM register set, when the device is
	 * created the caller can assign up to 4 memory regions.
	 */
	struct resource*	sc_mem_res[4];
	struct mtx		sc_mtx;
};

static struct omap_prcm_softc *g_omap_prcm_softc = NULL;

/**
 *	g_omap_prcm_spec - Resource specification used when allocating resources
 *
 *	This driver module can have up to four independant memory regions.
 */
static struct resource_spec g_omap_prcm_spec[] = {
	{ SYS_RES_MEMORY,   0,  RF_ACTIVE },
	{ SYS_RES_MEMORY,   1,  RF_ACTIVE | RF_OPTIONAL },
	{ SYS_RES_MEMORY,   2,  RF_ACTIVE | RF_OPTIONAL },
	{ SYS_RES_MEMORY,   3,  RF_ACTIVE | RF_OPTIONAL },
	{ -1,               0,  0 }
};


/**
 *	omap_clk_devmap - Array of clock devices, should be defined one per SoC 
 *
 *	This array is typically defined in one of the targeted omap??_prcm_clk.c
 *	files and is specific to the given SoC platform.  Each entry in the array
 *	corresponds to an individual clock device.
 */
extern const struct omap_clock_dev omap_clk_devmap[];



/**
 *	Macros for driver mutex locking
 */
#define OMAP_PRCM_LOCK(_sc)             mtx_lock(&(_sc)->sc_mtx)
#define	OMAP_PRCM_UNLOCK(_sc)           mtx_unlock(&(_sc)->sc_mtx)
#define OMAP_PRCM_LOCK_INIT(_sc) \
	mtx_init(&_sc->sc_mtx, device_get_nameunit(_sc->sc_dev), \
	         "omap_prcm", MTX_DEF)
#define OMAP_PRCM_LOCK_DESTROY(_sc)     mtx_destroy(&_sc->sc_mtx);
#define OMAP_PRCM_ASSERT_LOCKED(_sc)    mtx_assert(&_sc->sc_mtx, MA_OWNED);
#define OMAP_PRCM_ASSERT_UNLOCKED(_sc)  mtx_assert(&_sc->sc_mtx, MA_NOTOWNED);




/**
 *	omap_prcm_clk_dev - returns a pointer to the clock device with given id
 *	@clk: the ID of the clock device to get
 *
 *	Simply iterates through the clk_devmap global array and returns a pointer
 *	to the clock device if found. 
 *
 *	LOCKING:
 *	None
 *
 *	RETURNS:
 *	The pointer to the clock device on success, on failure NULL is returned.
 */
static const struct omap_clock_dev *
omap_prcm_clk_dev(struct omap_prcm_softc *sc, clk_ident_t clk)
{
	const struct omap_clock_dev *clk_dev;
	
	/* Find the clock within the devmap - it's a bit inefficent having a for 
	 * loop for this, but this function should only called when a driver is 
	 * being activated so IMHO not a big issue.
	 */
	clk_dev = &(omap_clk_devmap[0]);
	while (clk_dev->id != INVALID_CLK_IDENT) {
		if (clk_dev->id == clk) {
			return (clk_dev);
		}
		clk_dev++;
	}

	/* Sanity check we managed to find the clock */
	device_printf(sc->sc_dev, "Error: Failed to find clock device (%d)\n", clk);
	return (NULL);
}






/**
 *	omap_prcm_clk_valid - enables a clock for a particular module
 *	@clk: identifier for the module to enable, see omap_prcm.h for a list
 *	      of possible modules.
 *	         Example: OMAP3_MODULE_MMC1_ICLK or OMAP3_MODULE_GPTIMER10_FCLK.
 *	
 *	This function can enable either a functional or interface clock.
 *
 *	The real work done to enable the clock is really done in the callback
 *	function associated with the clock, this function is simply a wrapper
 *	around that.
 *
 *	LOCKING:
 *	Internally locks the driver context.
 *
 *	RETURNS:
 *	Returns 0 on success or positive error code on failure.
 */
int
omap_prcm_clk_valid(clk_ident_t clk)
{
	struct omap_prcm_softc *sc = g_omap_prcm_softc;
	int ret = 0;

	/* Sanity check */
	if (sc == NULL) {
		device_printf(sc->sc_dev, "Error: PRCM module not setup (%s)\n", __func__);
		return (ENOMEM);
	}

	OMAP_PRCM_LOCK(sc);

	if (omap_prcm_clk_dev(sc, clk) == NULL)
		ret = EINVAL;
	
	OMAP_PRCM_UNLOCK(sc);
	
	return (ret);
}


/**
 *	omap_prcm_clk_enable - enables a clock for a particular module
 *	@clk: identifier for the module to enable, see omap_prcm.h for a list
 *	      of possible modules.
 *	         Example: OMAP3_MODULE_MMC1_ICLK or OMAP3_MODULE_GPTIMER10_FCLK.
 *	
 *	This function can enable either a functional or interface clock.
 *
 *	The real work done to enable the clock is really done in the callback
 *	function associated with the clock, this function is simply a wrapper
 *	around that.
 *
 *	LOCKING:
 *	Internally locks the driver context.
 *
 *	RETURNS:
 *	Returns 0 on success or positive error code on failure.
 */
int
omap_prcm_clk_enable(clk_ident_t clk)
{
	struct omap_prcm_softc *sc = g_omap_prcm_softc;
	const struct omap_clock_dev *clk_dev;
	int ret;

	/* Sanity check */
	if (sc == NULL) {
		device_printf(sc->sc_dev, "Error: PRCM module not setup (%s)\n", __func__);
		return (ENOMEM);
	}

	OMAP_PRCM_LOCK(sc);

	/* Find the clock within the devmap - it's a bit inefficent having a for 
	 * loop for this, but this function should only called when a driver is 
	 * being activated so IMHO not a big issue.
	 */
	clk_dev = omap_prcm_clk_dev(sc, clk);

	/* Sanity check we managed to find the clock */
	if (clk_dev == NULL) {
		OMAP_PRCM_UNLOCK(sc);
		return (EINVAL);
	}

	/* Activate the clock */
	if (clk_dev->clk_activate)
		ret = clk_dev->clk_activate(clk_dev, sc->sc_mem_res);
	else
		ret = EINVAL;

	
	OMAP_PRCM_UNLOCK(sc);
	
	return (ret);
}


/**
 *	omap_prcm_clk_disable - disables a clock for a particular module
 *	@clk: identifier for the module to enable, see omap_prcm.h for a list
 *	      of possible modules.
 *	         Example: OMAP3_MODULE_MMC1_ICLK or OMAP3_MODULE_GPTIMER10_FCLK.
 *	
 *	This function can enable either a functional or interface clock.
 *
 *	The real work done to enable the clock is really done in the callback
 *	function associated with the clock, this function is simply a wrapper
 *	around that.
 *
 *	LOCKING:
 *	Internally locks the driver context.
 *
 *	RETURNS:
 *	Returns 0 on success or positive error code on failure.
 */
int
omap_prcm_clk_disable(clk_ident_t clk)
{
	struct omap_prcm_softc *sc = g_omap_prcm_softc;
	const struct omap_clock_dev *clk_dev;
	int ret;

	/* Sanity check */
	if (sc == NULL) {
		device_printf(sc->sc_dev, "Error: PRCM module not setup (%s)\n", __func__);
		return (ENOMEM);
	}

	OMAP_PRCM_LOCK(sc);

	/* Find the clock within the devmap - it's a bit inefficent having a for 
	 * loop for this, but this function should only called when a driver is 
	 * being activated so IMHO not a big issue.
	 */
	clk_dev = omap_prcm_clk_dev(sc, clk);

	/* Sanity check we managed to find the clock */
	if (clk_dev == NULL) {
		OMAP_PRCM_UNLOCK(sc);
		return (EINVAL);
	}

	/* Activate the clock */
	if (clk_dev->clk_deactivate)
		ret = clk_dev->clk_deactivate(clk_dev, sc->sc_mem_res);
	else
		ret = EINVAL;

	
	OMAP_PRCM_UNLOCK(sc);
	
	return (ret);
}

/**
 *	omap_prcm_clk_set_source - sets the source 
 *	@clk: identifier for the module to enable, see omap_prcm.h for a list
 *	      of possible modules.
 *	         Example: OMAP3_MODULE_MMC1_ICLK or OMAP3_MODULE_GPTIMER10_FCLK.
 *	
 *	This function can enable either a functional or interface clock.
 *
 *	The real work done to enable the clock is really done in the callback
 *	function associated with the clock, this function is simply a wrapper
 *	around that.
 *
 *	LOCKING:
 *	Internally locks the driver context.
 *
 *	RETURNS:
 *	Returns 0 on success or positive error code on failure.
 */
int
omap_prcm_clk_set_source(clk_ident_t clk, clk_src_t clksrc)
{
	struct omap_prcm_softc *sc = g_omap_prcm_softc;
	const struct omap_clock_dev *clk_dev;
	int ret;

	/* Sanity check */
	if (sc == NULL) {
		device_printf(sc->sc_dev, "Error: PRCM module not setup (%s)\n", __func__);
		return (ENOMEM);
	}

	OMAP_PRCM_LOCK(sc);

	/* Find the clock within the devmap - it's a bit inefficent having a for 
	 * loop for this, but this function should only called when a driver is 
	 * being activated so IMHO not a big issue.
	 */
	clk_dev = omap_prcm_clk_dev(sc, clk);

	/* Sanity check we managed to find the clock */
	if (clk_dev == NULL) {
		OMAP_PRCM_UNLOCK(sc);
		return (EINVAL);
	}

	/* Activate the clock */
	if (clk_dev->clk_set_source)
		ret = clk_dev->clk_set_source(clk_dev, clksrc, sc->sc_mem_res);
	else
		ret = EINVAL;

	
	OMAP_PRCM_UNLOCK(sc);
	
	return (ret);
}


/**
 *	omap_prcm_clk_get_source_freq - gets the source clock frequency
 *	@clk: identifier for the module to enable, see omap_prcm.h for a list
 *	      of possible modules.
 *	@freq: pointer to an integer that upon return will contain the src freq
 *	
 *	This function returns the frequency of the source clock.
 *
 *	The real work done to enable the clock is really done in the callback
 *	function associated with the clock, this function is simply a wrapper
 *	around that.
 *
 *	LOCKING:
 *	Internally locks the driver context.
 *
 *	RETURNS:
 *	Returns 0 on success or positive error code on failure.
 */
int
omap_prcm_clk_get_source_freq(clk_ident_t clk, unsigned int *freq)
{
	struct omap_prcm_softc *sc = g_omap_prcm_softc;
	const struct omap_clock_dev *clk_dev;
	int ret;

	/* Sanity check */
	if (sc == NULL) {
		device_printf(sc->sc_dev, "Error: PRCM module not setup (%s)\n", __func__);
		return (ENOMEM);
	}

	OMAP_PRCM_LOCK(sc);

	/* Find the clock within the devmap - it's a bit inefficent having a for 
	 * loop for this, but this function should only called when a driver is 
	 * being activated so IMHO not a big issue.
	 */
	clk_dev = omap_prcm_clk_dev(sc, clk);

	/* Sanity check we managed to find the clock */
	if (clk_dev == NULL) {
		OMAP_PRCM_UNLOCK(sc);
		return (EINVAL);
	}

	/* Get the source frequency of the clock */
	if (clk_dev->clk_get_source_freq)
		ret = clk_dev->clk_get_source_freq(clk_dev, freq, sc->sc_mem_res);
	else
		ret = EINVAL;

	
	OMAP_PRCM_UNLOCK(sc);
	
	return (ret);
}



/**
 *	omap_prcm_probe - probe function for the driver
 *	@dev: prcm device handle
 *
 *	Simply sets the name of the driver module.
 *
 *	LOCKING:
 *	None
 *
 *	RETURNS:
 *	Always returns 0
 */
static int
omap_prcm_probe(device_t dev)
{
	device_set_desc(dev, "TI OMAP Power, Reset and Clock Management");
	return (0);
}

/**
 *	omap_prcm_attach - attach function for the driver
 *	@dev: prcm device handle
 *
 *	Allocates and sets up the driver context, this simply entails creating a
 *	bus mappings for the PRCM register set.
 *
 *	LOCKING:
 *	None
 *
 *	RETURNS:
 *	Always returns 0
 */
static int
omap_prcm_attach(device_t dev)
{
	struct omap_prcm_softc *sc = device_get_softc(dev);
	int	err;
	
	if (g_omap_prcm_softc != NULL)
		return (ENXIO);
	
	sc->sc_dev = dev;
	
	OMAP_PRCM_LOCK_INIT(sc);
	
	/* Attempt to get the memory resource(s) */
	err = bus_alloc_resources(dev, g_omap_prcm_spec, sc->sc_mem_res);
	if (err) {
		device_printf(dev, "Error: could not allocate resources\n");
		return (ENXIO);
	}

	
	/* Setup DPLL5, this is the 120MHz clock used by the USB module */
	// omap3_prcm_setup_dpll5(sc, 120, 12);
	
	g_omap_prcm_softc = sc;

	return (0);
}

#define PRM_RSTCTRL		0x1b00
#define PRM_RSTCTRL_RESET	0x2

void
omap_prcm_reset(void)
{
	struct omap_prcm_softc *sc = g_omap_prcm_softc;
	bus_write_4(sc->sc_mem_res[0], PRM_RSTCTRL,
	    bus_read_4(sc->sc_mem_res[0], PRM_RSTCTRL) | PRM_RSTCTRL_RESET);
	bus_read_4(sc->sc_mem_res[0], PRM_RSTCTRL);
}

static device_method_t g_omap_prcm_methods[] = {
	DEVMETHOD(device_probe, omap_prcm_probe),
	DEVMETHOD(device_attach, omap_prcm_attach),
	{0, 0},
};

static driver_t g_omap_prcm_driver = {
	"omap_prcm",
	g_omap_prcm_methods,
	sizeof(struct omap_prcm_softc),
};
static devclass_t g_omap_prcm_devclass;

DRIVER_MODULE(omap_prcm, omap, g_omap_prcm_driver, g_omap_prcm_devclass, 0, 0);
MODULE_VERSION(omap_prcm, 1);
